import Vapor
import Fluent

struct AcronymsController: RouteCollection {
    func boot(routes: RoutesBuilder) throws {
        let acronymsRoute = routes.grouped("api", "acronyms")
        acronymsRoute.get(use: getAllHandler)
        acronymsRoute.post(use: createHandler)
        acronymsRoute.get(":acronymID", use: getHandler)
        acronymsRoute.delete(":acronymID", use: deleteHandler)
        acronymsRoute.put(":acronymID", use: updateHandler)
        acronymsRoute.get("search", use: searchHandler)
        acronymsRoute.get("first", use: firstHandler)
        acronymsRoute.get("sorted", use: sortedHandler)
        acronymsRoute.get(":acronymID", "user", use: getUserHandler)
        acronymsRoute.post(":acronymID", "categories", ":categoryID", use: attachCategoriesHandler)
        acronymsRoute.delete(":acronymID", "categories", ":categoryID", use: detachCategoriesHandler)
        acronymsRoute.get(":acronymID", "categories", use: getCategoriesHandler)
    }

    @Sendable
    func getAllHandler(_ req: Request) async throws -> [Acronym] {
        try await Acronym.query(on: req.db).all()
    }

    @Sendable
    func createHandler(_ req: Request) async throws -> Acronym {
        let data = try req.content.decode(AcronymCreateData.self)
        let acronym = Acronym(
            short: data.short, 
            long: data.long, 
            userID: data.userID
        )
        try await acronym.save(on: req.db)
        return acronym
    }

    @Sendable
    func getHandler(_ req: Request) async throws -> Acronym {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        return acronym
    }
    
    @Sendable
    func deleteHandler(_ req: Request) async throws -> HTTPStatus {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.delete(on: req.db)
        return .noContent
    }

    @Sendable
    func updateHandler(_ req: Request) async throws -> Acronym {
        let updatedAcronym = try req.content.decode(AcronymCreateData.self)
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        acronym.short = updatedAcronym.short
        acronym.long = updatedAcronym.long
        try await acronym.save(on: req.db)
        return acronym
    }

    @Sendable
    func searchHandler(_ req: Request) async throws -> [Acronym] {
        guard let searchTerm = req.query[String.self, at: "term"] else {
            throw Abort(.badRequest)
        }
        return try await Acronym.query(on: req.db).group(.or) { or in
            or.filter(\.$short == searchTerm)
            or.filter(\.$long == searchTerm)
        }.all()
    }

    @Sendable
    func firstHandler(_ req: Request) async throws -> Acronym {
        guard let firstAcronym = try await Acronym.query(on: req.db).first() else {
            throw Abort(.notFound)
        }
        return firstAcronym
    }

    @Sendable
    func sortedHandler(_ req: Request) async throws -> [Acronym] {
        try await Acronym.query(on: req.db).sort(\.$short, .ascending).all()
    }

    @Sendable
    func getUserHandler(_ req: Request) async throws -> User {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        return try await acronym.$user.get(on: req.db)
    }

    @Sendable
    func attachCategoriesHandler(_ req: Request) async throws -> HTTPStatus {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        guard let category = try await Category.find(req.parameters.get("categoryID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.$categories.attach(category, on: req.db)
        return .created
    }

    @Sendable
    func detachCategoriesHandler(_ req: Request) async throws -> HTTPStatus {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        guard let category = try await Category.find(req.parameters.get("categoryID"), on: req.db) else {
            throw Abort(.notFound)
        }
        try await acronym.$categories.detach(category, on: req.db)
        return .noContent
    }

    @Sendable
    func getCategoriesHandler(_ req: Request) async throws -> [Category] {
        guard let acronym = try await Acronym.find(req.parameters.get("acronymID"), on: req.db) else {
            throw Abort(.notFound)
        }
        return try await acronym.$categories.get(on: req.db)
    }
}

struct AcronymCreateData: Content {
    let short: String
    let long: String
    let userID: UUID
}