import Fluent
import XCTVapor

@testable import App

final class AcronymTests: XCTestCase {
    let acronymsURI = "/api/acronyms/"
    let acronymShort = "OMG"
    let acronymLong = "Oh My God"
    var app: Application!

    override func setUp() async throws {
        self.app = try await Application.make(.testing)
        try await configure(app)
        try await app.autoMigrate()
    }

    override func tearDown() async throws {
        try await app.autoRevert()
        try await self.app.asyncShutdown()
        self.app = nil
    }

    func testAcronymsCanBeRetrievedFromAPI() async throws {
        // 创建测试数据
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            on: app.db)

        // 测试 GET 请求
        try await app.test(
            .GET, acronymsURI,
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let acronyms = try res.content.decode([Acronym].self)
                XCTAssertEqual(acronyms.count, 1)
                XCTAssertEqual(acronyms[0].short, acronymShort)
                XCTAssertEqual(acronyms[0].long, acronymLong)
                XCTAssertEqual(acronyms[0].id, acronym.id)
            })
    }

    func testCreateAcronym() async throws {
        let user = try await User.create(on: app.db)
        let createData = AcronymCreateData(
            short: acronymShort,
            long: acronymLong,
            userID: user.id!)

        try await app.test(
            .POST, acronymsURI,
            beforeRequest: { req async throws in
                try req.content.encode(createData)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)

                let received = try res.content.decode(Acronym.self)
                XCTAssertEqual(received.short, acronymShort)
                XCTAssertEqual(received.long, acronymLong)
                XCTAssertNotNil(received.id)

                // 验证数据是否保存到数据库
                let savedAcronym = try await Acronym.find(received.id, on: app.db)
                XCTAssertNotNil(savedAcronym)
                XCTAssertEqual(savedAcronym?.short, acronymShort)
            })
    }

    func testDeleteAcronym() async throws {
        let acronym = try await Acronym.create(on: app.db)

        try await app.test(.DELETE, "\(acronymsURI)\(acronym.id!)") { res async throws in
            XCTAssertEqual(res.status, .noContent)

            // 验证数据是否从数据库中删除
            let deletedAcronym = try await Acronym.find(acronym.id, on: app.db)
            XCTAssertNil(deletedAcronym)
        }
    }

    func testGetAcronymByID() async throws {
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            on: app.db)

        try await app.test(.GET, "\(acronymsURI)\(acronym.id!)") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let receivedAcronym = try res.content.decode(Acronym.self)
            XCTAssertEqual(receivedAcronym.id, acronym.id)
            XCTAssertEqual(receivedAcronym.short, acronymShort)
            XCTAssertEqual(receivedAcronym.long, acronymLong)
        }
    }

    func testUpdateAcronym() async throws {
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            on: app.db)
        let updatedData = AcronymCreateData(
            short: "BRB",
            long: "Be Right Back",
            userID: acronym.$user.id)

        try await app.test(
            .PUT, "\(acronymsURI)\(acronym.id!)",
            beforeRequest: { req async throws in
                try req.content.encode(updatedData)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let updatedAcronym = try res.content.decode(Acronym.self)
                XCTAssertEqual(updatedAcronym.short, "BRB")
                XCTAssertEqual(updatedAcronym.long, "Be Right Back")
            })
    }

    func testSearchAcronym() async throws {
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            on: app.db)

        try await app.test(.GET, "\(acronymsURI)search?term=\(acronymShort)") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let acronyms = try res.content.decode([Acronym].self)
            XCTAssertEqual(acronyms.count, 1)
            XCTAssertEqual(acronyms[0].id, acronym.id)
            XCTAssertEqual(acronyms[0].short, acronymShort)
            XCTAssertEqual(acronyms[0].long, acronymLong)
        }
    }

    func testGetFirstAcronym() async throws {
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            on: app.db)

        try await app.test(.GET, "\(acronymsURI)first") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let firstAcronym = try res.content.decode(Acronym.self)
            XCTAssertEqual(firstAcronym.id, acronym.id)
            XCTAssertEqual(firstAcronym.short, acronymShort)
            XCTAssertEqual(firstAcronym.long, acronymLong)
        }
    }

    func testGetSortedAcronyms() async throws {
        let acronym1 = try await Acronym.create(
            short: "AAA",
            long: "First Acronym",
            on: app.db)
        let acronym2 = try await Acronym.create(
            short: "BBB",
            long: "Second Acronym",
            on: app.db)

        try await app.test(.GET, "\(acronymsURI)sorted") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let acronyms = try res.content.decode([Acronym].self)
            XCTAssertEqual(acronyms.count, 2)
            XCTAssertEqual(acronyms[0].id, acronym1.id)
            XCTAssertEqual(acronyms[1].id, acronym2.id)
        }
    }
}

extension AcronymTests {
    func testGetAcronymUser() async throws {
        let user = try await User.create(
            name: "Alice",
            username: "alice",
            on: app.db)
        let acronym = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            user: user,
            on: app.db)

        try await app.test(.GET, "\(acronymsURI)\(acronym.id!)/user") { res async throws in
            let acronymUser = try res.content.decode(User.self)
            XCTAssertEqual(acronymUser.id, user.id)
            XCTAssertEqual(acronymUser.name, "Alice")
            XCTAssertEqual(acronymUser.username, "alice")
        }
    }

    func testGetUserAcronyms() async throws {
        let user = try await User.create(on: app.db)
        let acronym1 = try await Acronym.create(
            short: acronymShort,
            long: acronymLong,
            user: user,
            on: app.db)
        _ = try await Acronym.create(
            short: "LOL",
            long: "Laugh Out Loud",
            user: user,
            on: app.db)

        try await app.test(.GET, "/api/users/\(user.id!)/acronyms") { res async throws in
            let acronyms = try res.content.decode([Acronym].self)
            XCTAssertEqual(acronyms.count, 2)
            XCTAssertEqual(acronyms[0].id, acronym1.id)
            XCTAssertEqual(acronyms[0].short, acronymShort)
            XCTAssertEqual(acronyms[0].long, acronymLong)
        }
    }

    func testAddCategoryToAcronym() async throws {
        let acronym = try await Acronym.create(on: app.db)
        let category = try await Category.create(
            name: "Technology",
            on: app.db)
        
        try await app.test(
            .POST,
            "\(acronymsURI)\(acronym.id!)/categories/\(category.id!)",
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .created)
                
                // 验证关联是否创建成功
                let categories = try await acronym.$categories.get(on: app.db)
                XCTAssertEqual(categories.count, 1)
                XCTAssertEqual(categories[0].id, category.id)
                XCTAssertEqual(categories[0].name, "Technology")
            })
    }
    
    func testRemoveCategoryFromAcronym() async throws {
        let acronym = try await Acronym.create(on: app.db)
        let category = try await Category.create(
            name: "Technology",
            on: app.db)
        try await acronym.$categories.attach(category, on: app.db)
        
        try await app.test(
            .DELETE,
            "\(acronymsURI)\(acronym.id!)/categories/\(category.id!)",
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .noContent)
                
                // 验证关联是否被删除
                let categories = try await acronym.$categories.get(on: app.db)
                XCTAssertEqual(categories.count, 0)
            })
    }
}


