
import Fluent
import XCTVapor

@testable import App

final class CategoryTests: XCTestCase {
    let categoriesURI = "/api/categories/"
    var app: Application!

    override func setUp() async throws {
        self.app = try await Application.make(.testing)
        try await configure(app)
        try await app.autoMigrate()
    }

    override func tearDown() async throws {
        try await app.autoRevert()
        try await self.app.asyncShutdown()
        self.app = nil
    }

    func testCreateCategory() async throws {
        let category = Category(name: "Technology")

        try await app.test(
            .POST, categoriesURI,
            beforeRequest: { req async throws in
                try req.content.encode(category)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let receivedCategory = try res.content.decode(Category.self)
                XCTAssertEqual(receivedCategory.name, category.name)
                XCTAssertNotNil(receivedCategory.id)

                let savedCategory = try await Category.find(receivedCategory.id, on: app.db)
                XCTAssertNotNil(savedCategory)
                XCTAssertEqual(savedCategory?.name, category.name)
            })
    }

    func testDeleteCategory() async throws {
        let category = try await Category.create(name: "Technology", on: app.db)

        try await app.test(.DELETE, "\(categoriesURI)\(category.id!)") { res async throws in
            XCTAssertEqual(res.status, .noContent)

            let deletedCategory = try await Category.find(category.id, on: app.db)
            XCTAssertNil(deletedCategory)
        }
    }

    func testGetCategoryByID() async throws {
        let category = try await Category.create(name: "Technology", on: app.db)

        try await app.test(.GET, "\(categoriesURI)\(category.id!)") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let receivedCategory = try res.content.decode(Category.self)
            XCTAssertEqual(receivedCategory.id, category.id)
            XCTAssertEqual(receivedCategory.name, category.name)
        }
    }

    func testUpdateCategory() async throws {
        let category = try await Category.create(name: "Technology", on: app.db)
        let updatedCategory = Category(name: "Science")

        try await app.test(
            .PUT, "\(categoriesURI)\(category.id!)",
            beforeRequest: { req async throws in
                try req.content.encode(updatedCategory)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let receivedCategory = try res.content.decode(Category.self)
                XCTAssertEqual(receivedCategory.name, "Science")
            })
    }

    func testGetCategoryAcronyms() async throws {
        let category = try await Category.create(name: "Technology", on: app.db)
        let acronym1 = try await Acronym.create(
            short: "OMG",
            long: "Oh My God",
            on: app.db)
        let acronym2 = try await Acronym.create(
            short: "LOL",
            long: "Laugh Out Loud",
            on: app.db)
        try await category.$acronyms.attach(acronym1, on: app.db)
        try await category.$acronyms.attach(acronym2, on: app.db)

        try await app.test(.GET, "\(categoriesURI)\(category.id!)/acronyms") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let acronyms = try res.content.decode([Acronym].self)
            XCTAssertEqual(acronyms.count, 2)
            XCTAssertEqual(acronyms[0].id, acronym1.id)
            XCTAssertEqual(acronyms[1].id, acronym2.id)
        }
    }
}