import Fluent
import XCTVapor

@testable import App

final class UserTests: XCTestCase {
    let usersURI = "/api/users/"
    var app: Application!

    override func setUp() async throws {
        self.app = try await Application.make(.testing)
        try await configure(app)
        try await app.autoMigrate()
    }

    override func tearDown() async throws {
        try await app.autoRevert()
        try await self.app.asyncShutdown()
        self.app = nil
    }

    func testUsersCanBeRetrievedFromAPI() async throws {
        let user = try await User.create(on: app.db)

        try await app.test(.GET, usersURI) { res async throws in
            XCTAssertEqual(res.status, .ok)
            let users = try res.content.decode([User].self)
            XCTAssertEqual(users.count, 1)
            XCTAssertEqual(users[0].name, user.name)
            XCTAssertEqual(users[0].username, user.username)
            XCTAssertEqual(users[0].id, user.id)
        }
    }

    func testCreateUser() async throws {
        let user = User(name: "Alice", username: "alice")

        try await app.test(
            .POST, usersURI,
            beforeRequest: { req async throws in
                try req.content.encode(user)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let receivedUser = try res.content.decode(User.self)
                XCTAssertEqual(receivedUser.name, user.name)
                XCTAssertEqual(receivedUser.username, user.username)
                XCTAssertNotNil(receivedUser.id)

                let savedUser = try await User.find(receivedUser.id, on: app.db)
                XCTAssertNotNil(savedUser)
                XCTAssertEqual(savedUser?.name, user.name)
                XCTAssertEqual(savedUser?.username, user.username)
            })
    }

    func testDeleteUser() async throws {
        let user = try await User.create(on: app.db)

        try await app.test(.DELETE, "\(usersURI)\(user.id!)") { res async throws in
            XCTAssertEqual(res.status, .noContent)

            let deletedUser = try await User.find(user.id, on: app.db)
            XCTAssertNil(deletedUser)
        }
    }

    func testGetUserByID() async throws {
        let user = try await User.create(on: app.db)

        try await app.test(.GET, "\(usersURI)\(user.id!)") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let receivedUser = try res.content.decode(User.self)
            XCTAssertEqual(receivedUser.id, user.id)
            XCTAssertEqual(receivedUser.name, user.name)
            XCTAssertEqual(receivedUser.username, user.username)
        }
    }

    func testUpdateUser() async throws {
        let user = try await User.create(on: app.db)
        let updatedUser = User(name: "Bob", username: "bob")

        try await app.test(
            .PUT, "\(usersURI)\(user.id!)",
            beforeRequest: { req async throws in
                try req.content.encode(updatedUser)
            },
            afterResponse: { res async throws in
                XCTAssertEqual(res.status, .ok)
                let receivedUser = try res.content.decode(User.self)
                XCTAssertEqual(receivedUser.name, "Bob")
                XCTAssertEqual(receivedUser.username, "bob")
            })
    }

    func testGetUserAcronyms() async throws {
        let user = try await User.create(on: app.db)
        let acronym1 = try await Acronym.create(
            short: "OMG",
            long: "Oh My God",
            user: user,
            on: app.db)
        _ = try await Acronym.create(
            short: "LOL",
            long: "Laugh Out Loud",
            user: user,
            on: app.db)

        try await app.test(.GET, "\(usersURI)\(user.id!)/acronyms") { res async throws in
            XCTAssertEqual(res.status, .ok)
            let acronyms = try res.content.decode([Acronym].self)
            XCTAssertEqual(acronyms.count, 2)
            XCTAssertEqual(acronyms[0].id, acronym1.id)
            XCTAssertEqual(acronyms[0].short, "OMG")
            XCTAssertEqual(acronyms[0].long, "Oh My God")
        }
    }
}
